//////////////////////////////////////////////////////////////////////////
// Spuzzum's Trigonometric Library           by Jeremy "Spuzzum" Gibson //
//////////////////////////////////////////////////////////////////////////
//
//   A set of standard trigonometric functions, based on binomial       //
// series, as suggested by Al.  Includes optional procs for generating  //
// numbers that are more highly accurate, if you need to.  However,     //
// as Al said, the quick procs qsin(), qcos(), and qtan() should be     //
// sufficient for most purposes, as you will probably call them a lot.  //
//                                                                      //
//////////////////////////////////////////////////////////////////////////
// Expanded Definitions:                                                //
// sin -> sine            qsin -> quick sine                            //
// asin -> arc sine       qasin -> quick arc sine                       //
// tan -> tangent         qtan -> quick tangent                         //
// atan -> arc tangent    qatan -> quick arc tangent                    //
// cos -> cosine          qcos -> quick cosine                          //
// acos -> arc cosine     qacos -> quick arc cosine                     //
//////////////////////////////////////////////////////////////////////////
// Recommendations / Caveats:                                           //
//   In the interest of speed, you should immediately opt for the quick //
// versions of these procs.  They give must faster results, taxing the  //
// server much less and providing you with much steadier output.        //
//   If for some reason, you need results that are more specific (if    //
// you are calculating vectors over long distances, to name one), you   //
// may use the iterative versions of the quick procs.  These default to //
// the same number of iterations as the quicker versions, to ensure you //
// don't make a mistake and forget to specify the amount.  If you do    //
// use the slow versions, in the interest of speed, don't go any higher //
// than 5 or 6 iterations.  Otherwise you'll receive exceedingly large  //
// computations for exceedingly meaningless differences.  In iterative  //
// tan, you shouldn't go much higher than 4 iterations.                 //
//////////////////////////////////////////////////////////////////////////

/*
Proc Overview:
- sin(N as num, iterations=3) //Calculates sine
  Takes a number in radians.  Returns the sine of that number.  Calculates
sine accurately through a number of iterations.  Iterations is optional.


*********
- asin(N as num, iterations=3) //Calculates arc sine (inverse sine)
  Takes a computed sine.  Returns the number in radians.  Calculates
accurately through a number of iterations.  Iterations is optional.

- tan(N as num, iterations=2) //Calculates tangent
  Takes a number in radians.  Returns the tangent of that number.

*Because I haven't taken number theory to the point of determining power
series, I am unable to provide iterative functions to do these tasks.
Perhaps someone with greater than grade 11 education (hey, I'm only 15,
after all) could elucidate?*
*********

- atan(N as num, iterations=2) //Calculates arc tangent (inverse tangent)
  Takes a computed tangent.  Returns the number in radians.  Calculates
accurately through a number of iterations.  Iterations is optional.

- cos(N as num, iterations=3) //Calculates cosine
  Takes a number in radians.  Returns the cosine of that number.
Calculates cosine accurately through a number of iterations.  Iterations
is optional.

- acos(N as num, iterations=3) //Calculates arc cosine (inverse cosine)
  Takes a computed cosine.  Returns the number in radians.  Calculates
accurately through a number of iterations.  Iterations is optional.


- qsin(N as num) //Calculates sine
  Takes a number in radians.  Returns the sine of that number.  Used when
speed is more important than puritan accuracy.

- qtan(N as num) //Calculates tangent
  Takes a number in radians.  Returns the sine of that number.  Used when
speed is more important than puritan accuracy.

- qcos(N as num) //Calculates cosine
  Takes a number in radians.  Returns the sine of that number.  Used when
speed is more important than puritan accuracy.

- qasin(N as num) //Calculates arc sine (inverse sine)
  Takes a number in radians.  Returns the sine of that number.  Used when
speed is more important than puritan accuracy.

- qatan(N as num) //Calculates arc tangent (inverse tan)
  Takes a number in radians.  Returns the sine of that number.  Used when
speed is more important than puritan accuracy.

- qacos(N as num) //Calculates arc cosine (inverse cosine)
  Takes a number in radians.  Returns the sine of that number.  Used when
speed is more important than puritan accuracy.


- factorial(N as num) //gives the factorial of a number (3! = 1*2*3 = 6)
  Computes the factorial of the specified argument.

*/

var
	const
		pi = 3.1416 //BYOND is only accurate to 4 decimal places


proc
	factorial(N as num)
		if(N<=0) return 1
		return .(N-1)*N


	sin(N as num,iterations=3)
		var
			num = N
			exponent = 3
			i = 1
		while(i++ <= iterations)
			if(iterations%2 == 0)
				num += (N**exponent)/factorial(exponent)
			else
				num -= (N**exponent)/factorial(exponent)
			exponent += 2
		return num

	cos(N as num,iterations=3)
		var
			num = 1
			exponent = 2
			i = 1
		while(i++ <= iterations)
			if(iterations%2 == 0)
				num += (N**exponent)/factorial(exponent)
			else
				num -= (N**exponent)/factorial(exponent)

	acos(N as num,iterations=3)


	atan(N as num,iterations=3)


	qsin(N as num) //3 iterations
		var/num = N
		num -= (N**3)/6
		num += (N**5)/120
		num -= (N**7)/5040
		return num

	qcos(N as num) //3 iterations
		var/num = 1
		num -= (N**2)/2
		num += (N**4)/24
		num -= (N**6)/720
		return num

	qtan(N as num) //2 iterations
		var/num = N
		num += (N**3)/3
		num += 2*(N**5)/15
		return num

	qasin(N as num)
		var/num = N
		num += 0.5*((N**3)/3)
		num += (3/8)*((N**5)/5)
		num += (15/48)*((N**7)/7)
		return num

	qacos(N as num)
		return (pi/2)-qasin(N)

	qatan(N as num)
		var/num = N
		num -= (N**3)/3
		num += (N**5)/5
		num -= (N**7)/7
		return num